import { addToMaxChunkSize } from "../utils"
import { METLO_META } from "."
import { sendMetloTrace, shouldBlock } from "../rust_binds"
import { MetloConfig } from "../types"
import { RequestHandler, Request, Response } from "express-serve-static-core"
import { Logger } from "../utils/logger"

export interface MetloConfigExpress<
  _Req extends Request = Request,
  _Res extends Response = Response,
> extends MetloConfig {
  getUser?: (req: _Req, res: _Res) => string
  rejectResponse?: (req: _Req, res: _Res) => void
}

function versionCheck() {
  return true
}

const initialize = <
  _reqType extends Request = Request,
  _resType extends Response = Response,
>() => {
  if (!versionCheck()) {
    return
  }

  function compileRequest(_req: Request, _res: Response) {
    return {
      url: {
        host: (
          METLO_META.opts?.apiHost || _req.socket.remoteAddress || ""
        ).toString(),
        path: (_req.baseUrl || "" + _req.path || "").toString(),
        parameters: Object.entries(_req.query).map(([k, v]) => ({
          name: k.toString(),
          value: v.toString(),
        })),
      },
      headers: Object.entries(_req.headers).map(([k, v]) => ({
        name: k.toString(),
        value: v.toString(),
      })),
      body: (typeof _req.body === "object"
        ? JSON.stringify(_req.body)
        : _req.body || ""
      )
        .toString()
        .slice(0, METLO_META.maxBodyCaptureSize),
      method: _req.method.toString(),
      user: METLO_META.getUser
        ? (METLO_META as MetloConfigExpress).getUser(_req, _res)
        : null,
    }
  }

  function compileMeta(_req, _res) {
    return {
      environment: process.env.NODE_ENV || "development",
      incoming: true,
      source: _req.socket.remoteAddress.toString(),
      sourcePort: _req.socket.remotePort,
      destination: _req.socket.localAddress.toString(),
      destinationPort: _req.socket.localPort,
      metloSource: "node/express",
    }
  }

  function compileResponse(_req, _res, responseBody) {
    return {
      status: _res.statusCode,
      headers: Object.entries(_res.getHeaders()).map(([k, v]) => ({
        name: k.toString(),
        value: v.toString(),
      })),
      body: (typeof responseBody === "object"
        ? JSON.stringify(responseBody)
        : responseBody || "{}"
      )
        .toString()
        .slice(0, METLO_META.maxBodyCaptureSize),
    }
  }

  function compileInformation(_req, _res, responseBody) {
    try {
      const data = JSON.stringify({
        request: compileRequest(_req, _res),
        response: compileResponse(_req, _res, responseBody),
        meta: compileMeta(_req, _res),
      })
      sendMetloTrace(data)
    } catch (e) {
      Logger().warn(`Metlo encountered an error ${e}`);
    }
  }

  if (METLO_META.metloEnabled) {
    let middleware: RequestHandler = (req, res, next) => {
      try {
        let shouldRejectChunks = false

        const oldWrite = res.write
        const oldEnd = res.end

        const chunks: Array<Buffer> = []

        const chunkParse = (chunk: Buffer | string) => {
          try {
            const currSize = chunks.reduce(
              (acc, _ch) => acc + _ch.byteLength,
              0,
            )
            if (typeof chunk === "string") {
              if (
                chunk &&
                METLO_META.maxBodyCaptureSize - currSize > 0 &&
                !shouldRejectChunks
              ) {
                addToMaxChunkSize(
                  chunks,
                  Buffer.from(chunk),
                  METLO_META.maxBodyCaptureSize,
                )
              } else {
                shouldRejectChunks = true
              }
            } else if (chunk instanceof Buffer) {
              if (
                chunk &&
                METLO_META.maxBodyCaptureSize - currSize > 0 &&
                !shouldRejectChunks
              ) {
                if (chunk) {
                  addToMaxChunkSize(
                    chunks,
                    chunk,
                    METLO_META.maxBodyCaptureSize,
                  )
                }
              } else {
                shouldRejectChunks = true
              }
            }
          } catch (err) {
            Logger().log("debug", err)
          }
        }

        res.write = function (chunk: Buffer | string) {
          chunkParse(chunk)
          return oldWrite.apply(res, arguments)
        }
        //@ts-ignore
        res.end = function (chunk: Buffer | string) {
          chunkParse(chunk)
          oldEnd.apply(res, arguments)
        }

        res.on("finish", () => {
          compileInformation(
            req,
            res,
            Buffer.concat(chunks).toString("utf8") || "{}",
          )
        })

        if (!METLO_META.disableBlocking) {
          try {
            const apiReq = JSON.stringify(compileRequest(req, res))
            const apiMeta = JSON.stringify(compileMeta(req, res))
            const willBlock = shouldBlock(apiReq, apiMeta)
            if (willBlock) {
              if (METLO_META.rejectResponse) {
                return (METLO_META as MetloConfigExpress).rejectResponse(req, res)
              } else {
                return res.status(403).send()
              }
            }
          } catch (e) {
            Logger().warn(`Metlo encountered an error ${e}`);
          }
        }
      } catch (err) {
        Logger().log("debug", err)
      }
      return next()
    }
    return middleware
  } else {
    return dummyModule()
  }
}

export const dummyModule = () => {
  return (req, res, next) => {
    next()
  }
}

export default initialize
