import { MAX_BODY_SIZE } from "../utils/constants"
import { startMetlo } from "../rust_binds"
import { MetloConfig } from "../types"
import { Logger } from "../utils/logger"
import { METLO_ENABLED } from "../utils"

export interface MetloConfigInternal extends MetloConfig {
  metloEnabled: boolean
}

export const METLO_META: MetloConfigInternal = {
  key: "",
  host: "",
  backendPort: -1,
  collectorPort: -1,
  opts: {},
  maxBodyCaptureSize: 5 * 1024,
  metloEnabled: true,
  disableBlocking: false,
}

export const initMiddleware = (props: MetloConfig) => {
  let shouldMetloBeEnabled = true

  METLO_META.logLevel = props.logLevel || "info"
  Logger(METLO_META.logLevel)

  try {
    new URL(props.host)
  } catch (err) {
    shouldMetloBeEnabled = false
    Logger().log("error", `Metlo Host is not a proper url : ${props.host}`)
  }
  METLO_META.host = props.host

  if (typeof props.key === "string") {
    METLO_META.key = props.key
  } else {
    shouldMetloBeEnabled = false
    Logger().log("error", "Metlo API key passed is not a valid string.")
  }

  if (props.opts) {
    METLO_META.opts = props.opts
  } else {
    METLO_META.opts = {}
  }

  if (props.backendPort) {
    if (
      typeof props.backendPort === "number" &&
      Number.isInteger(props.backendPort)
    ) {
      METLO_META.backendPort = props.backendPort
    } else {
      shouldMetloBeEnabled = false
      Logger().log("error", "Metlo backend port passed is not a valid number.")
    }
  } else {
    METLO_META.backendPort = props.host.includes("app.metlo.com") ? 443 : 8000
  }

  if (props.collectorPort) {
    if (
      typeof props.collectorPort === "number" &&
      Number.isInteger(props.collectorPort)
    ) {
      METLO_META.collectorPort = props.collectorPort
    } else {
      shouldMetloBeEnabled = false
      Logger().log("error", "Metlo backend port passed is not a valid number.")
    }
  } else {
    METLO_META.collectorPort = 8081
  }

  if (props.encryptionKey) {
    if (typeof props.encryptionKey === "string") {
      METLO_META.encryptionKey = props.encryptionKey
    } else {
      Logger().log(
        "error",
        "Metlo encryption key passed is not a valid string.",
      )
    }
  }

  if (props.getUser) {
    if (typeof props.getUser === "function") {
      METLO_META.getUser = props.getUser
    } else {
      shouldMetloBeEnabled = false
      Logger().log("error", "getUser passed is not a valid function.")
    }
  }

  if (props.rejectResponse) {
    if (typeof props.rejectResponse === "function") {
      METLO_META.rejectResponse = props.rejectResponse
    } else {
      shouldMetloBeEnabled = false
      Logger().log("error", "rejectResponse passed is not a valid function.")
    }
  }

  if (
    props.maxBodyCaptureSize &&
    typeof props.maxBodyCaptureSize === "number" &&
    Number.isInteger(props.maxBodyCaptureSize)
  ) {
    METLO_META.maxBodyCaptureSize = props.maxBodyCaptureSize
  } else {
    if (props.maxBodyCaptureSize) {
      if (typeof props.maxBodyCaptureSize !== "number") {
        Logger().log(
          "error",
          `maxBodyCaptureSize passed is not a valid number. Defaulting to size of ${MAX_BODY_SIZE} bytes`,
        )
      } else if (!Number.isInteger(props.maxBodyCaptureSize)) {
        Logger().log(
          "error",
          `maxBodyCaptureSize passed is not a valid integer. Defaulting to size of ${MAX_BODY_SIZE} bytes`,
        )
      }
    }
    METLO_META.maxBodyCaptureSize = MAX_BODY_SIZE
  }

  let disableBlocking = false
  if (typeof props.disableBlocking === "boolean") {
    disableBlocking = Boolean(props.disableBlocking)
    if (disableBlocking) {
      Logger().log("debug", "Request Blocking by Metlo has been disabled")
    }
  } else {
    let flagValue = process.env.METLO_DISABLE_BLOCKING
    if (
      flagValue !== undefined &&
      (flagValue === "true" || flagValue === "1" || flagValue === "yes")
    ) {
      disableBlocking = true
    }
  }
  METLO_META.disableBlocking = disableBlocking

  if (shouldMetloBeEnabled) {
    METLO_META.metloEnabled = METLO_ENABLED()
  } else {
    METLO_META.metloEnabled = false
  }

  if (METLO_META.metloEnabled) {
    startMetlo(
      METLO_META.host,
      METLO_META.key,
      METLO_META.backendPort,
      METLO_META.collectorPort,
      METLO_META.encryptionKey,
      METLO_META.logLevel,
    )
      .then(() => {
        Logger().log("debug", "Metlo has started")
      })
      .catch(err => {
        Logger().log("error", "Metlo encountered an error while initializing ")
        Logger().log("error", err)
      })
  } else {
    Logger().log("info", "Metlo is not enabled")
  }
}
export default initMiddleware
